from pathlib import Path

HEADER_FILE = Path.home() / '.FeelUOwn' / 'data' / 'ytmusic_header.json'
REQUIRED_COOKIE_FIELDS = ['HSID', 'SSID', 'APISID', 'SAPISID', '__Secure-3PAPISID', 'LOGIN_INFO', '__Secure-1PAPISID',
                          'SID', '__Secure-1PSID', '__Secure-3PSID', '__Secure-3PSIDCC']



from feeluown.app import App

__alias__ = 'ytmusic'
__version__ = '0.3.0'
__desc__ = 'YouTube Music plugin'


ui_mgr = None


def init_config(config):
    # For example: http://127.0.0.1:7890. This will be used in API and media accessing.
    config.deffield('HTTP_PROXY', type_=str, default='', desc='YouTube Music HTTP proxy')
    config.deffield('HTTP_TIMEOUT', type_=int, default=2, desc='HTTP requests timeout')


def enable(app: App):
    global ui_mgr

    from fuo_ytmusic.ui import YtmusicUiManager
    from fuo_ytmusic.provider import provider

    provider.setup_http_proxy(app.config.ytmusic.HTTP_PROXY)
    provider.setup_http_timeout(app.config.ytmusic.HTTP_TIMEOUT)
    app.library.register(provider)
    if app.mode & app.GuiMode:
        ui_mgr = ui_mgr or YtmusicUiManager(app)


def disable(app: App):
    global ui_mgr

    provider = app.library.get('ytmusic')
    if provider is not None:
        app.library.deregister(provider)
    ui_mgr = None



import asyncio
import logging
import json
from pathlib import Path

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel, QAction, QInputDialog, QMessageBox, QTextEdit
from feeluown.utils import aio
from feeluown.gui.widgets.login import CookiesLoginDialog
from feeluown.uimodels.my_music import MyMusicUiManager
from feeluown.uimodels.playlist import PlaylistUiManager
from feeluown.uimodels.provider import ProviderUiManager

from fuo_ytmusic.consts import HEADER_FILE, REQUIRED_COOKIE_FIELDS

# QML page has two problems
# 1. they may block the whole UI.
# 2. they consumes much memory.
# from .page_explore_qml import render as explore_render
# from .page_more import render as more_render
from .page_fav import render as fav_render
from .service import YtmusicPrivacyStatus


logger = logging.getLogger(__name__)


class YtmusicUiManager:
    def __init__(self, app):
        self._app = app
        self._provider = app.library.get('ytmusic')
        self._pvd_uimgr: ProviderUiManager = app.pvd_uimgr
        self._pvd_item = self._pvd_uimgr.create_item(
            name=self._provider.identifier,
            text=self._provider.name,
            desc=self._provider.name,
            colorful_svg=(Path(__file__).parent / 'assets' / 'icon.svg').as_posix()
        )
        self._pvd_item.clicked.connect(self.login_or_show)
        self._pvd_uimgr.add_item(self._pvd_item)
        self._app.browser.route('/providers/ytmusic/fav')(fav_render)
        # self._app.browser.route('/providers/ytmusic/explore')(explore_render)
        # self._app.browser.route('/providers/ytmusic/more')(more_render)

    def login_or_show(self):
        if self._provider.user is None:
            dialog = LoginDialog(self._provider)
            dialog.login_succeed.connect(lambda: asyncio.ensure_future(self.load_user()))
            dialog.open()
            dialog.autologin()
        else:
            asyncio.ensure_future(self.load_user())

    def new_playlist(self):
        name, o1 = QInputDialog.getText(self._app.ui.left_panel.playlists_header, '新建歌单', '请输入歌单名称:')
        if not o1:
            return
        desc, o2 = QInputDialog.getText(self._app.ui.left_panel.playlists_header, '新建歌单', '请输入歌单描述:')
        if not o2:
            return
        privacy_status, o3 = QInputDialog.getItem(self._app.ui.left_panel.playlists_header, '新建歌单', '请选择歌单权限:',
                                                  [i.value for i in YtmusicPrivacyStatus])
        if not o3:
            return
        ok = self._provider.create_playlist(name, desc, YtmusicPrivacyStatus.parse(privacy_status))
        if not ok:
            QMessageBox.warning(self._app.ui.left_panel.playlists_header, '新建歌单', '创建失败！')
        else:
            QMessageBox.information(self._app.ui.left_panel.playlists_header, '新建歌单', '创建成功！')

    async def load_user(self):
        user = self._provider.user
        self._app.ui.left_panel.playlists_con.show()
        self._app.ui.left_panel.my_music_con.show()

        # hack fuo to support add playlist
        pl_header: QLabel = self._app.ui.left_panel.playlists_header
        pl_header.setContextMenuPolicy(Qt.ActionsContextMenu)
        for a in pl_header.actions():
            pl_header.removeAction(a)
        new_pl_action = QAction('新建歌单', pl_header)
        pl_header.addAction(new_pl_action)
        new_pl_action.triggered.connect(self.new_playlist)

        mymusic_mgr: MyMusicUiManager = self._app.mymusic_uimgr
        playlists_mgr: PlaylistUiManager = self._app.pl_uimgr

        # explore_item = mymusic_mgr.create_item('🔮 发现音乐')
        my_fav_item = mymusic_mgr.create_item('⭐️ 收藏与关注')
        # more_item = mymusic_mgr.create_item('☁️ 上传的音乐')
        # explore_item.clicked.connect(lambda: self._app.browser.goto(page='/providers/ytmusic/explore'), weak=False)
        my_fav_item.clicked.connect(lambda: self._app.browser.goto(page='/providers/ytmusic/fav'), weak=False)
        # more_item.clicked.connect(lambda: self._app.browser.goto(page='/providers/ytmusic/more'), weak=False)
        mymusic_mgr.clear()
        # mymusic_mgr.add_item(explore_item)
        mymusic_mgr.add_item(my_fav_item)
        # mymusic_mgr.add_item(more_item)

        playlists_mgr.clear()
        self._pvd_item.text = f'{user.name} - 已登录'

        loop = asyncio.get_event_loop()
        pls = await loop.run_in_executor(None, self._provider.library_playlists)
        playlists_mgr.add(pls)


class LoginDialog(CookiesLoginDialog):
    def __init__(self, provider):
        super(LoginDialog, self).__init__(uri='https://music.youtube.com/',
                                          required_cookies_fields=REQUIRED_COOKIE_FIELDS)
        self._provider = provider

        self.auth_text_edit = QTextEdit(self)
        self.auth_text_edit.setAcceptRichText(False)
        self.auth_text_edit.setPlaceholderText(
            '请打开 music.youtube.com 并登陆，'
            '然后从浏览器中复制 HTTP 请求 header 中的 Authorization 字段：\n\n'
            '它类似这样：SAPISIDHASH 123333333_abfasdfs12fadcdfa2d'
        )
        self.__hint_label = QLabel(
            '还有另外一种登陆方式，你可以参考 ytmusicapi 的使用文档，使用'
            ' ytmusicapi oauth 命令来生成登陆所需信息。你如果登陆成功，'
            '它会生成一个 oauth.json 文件，你把 oauth.json 文件移动到'
            ' ~/.FeelUOwn/data/ytmusic_header.json ，然后重新点击图标登陆即可。'
        )
        self.__hint_label.setWordWrap(True)
        # weblogin can't fetch the authorization field, so diable it.
        self.weblogin_btn.hide()
        self._layout.insertWidget(0, self.auth_text_edit)
        self._layout.addWidget(self.__hint_label)
        self.setWindowFlags(self.windowFlags() | Qt.Dialog)

    def setup_user(self, user):
        self._provider.user = user

    def get_cookies(self):
        cookie = self.cookies_text_edit.toPlainText()
        auth = self.auth_text_edit.toPlainText()
        if not auth or not cookie:
            self.show_hint(f'authorization is empty, you must fill it', color='orange')
            return None

        agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:88.0) Gecko/20100101 Firefox/88.0'
        # Note ytmusicapi needs the entire header, not only the authorization and cookie parts.
        header = {
            "user-agent": agent,
            "accept": "*/*",
            "accept-encoding": "gzip, deflate",
            "content-type": "application/json",
            "content-encoding": "gzip",
            "origin": 'https://music.youtube.com',
            'authorization': auth,
            'cookie': cookie,
        }
        return header

    async def user_from_cookies(self, cookies):
        return self._provider.user_from_cookie(cookies)

    def load_user_cookies(self):
        self.load_header_file()

    def load_header_file(self):
        if HEADER_FILE.exists():
            with HEADER_FILE.open('r') as f:
                data = json.load(f)
                return data

    def autologin(self):
        """Overload super.autologin."""
        header_or_oauth = self.load_header_file()
        if header_or_oauth is not None:
            self.show_hint('正在尝试加载已有用户...', color='green')
            if 'cookie' in header_or_oauth:  # It is a header.
                cookie = header_or_oauth['cookie']
                auth = header_or_oauth['authorization']
                self.cookies_text_edit.setText(cookie)
                self.auth_text_edit.setText(auth)
            else:
                logger.debug('The header file is a oauth.json, will not load it to UI')
            aio.create_task(self.login_with_cookies(header_or_oauth))

    def dump_user_cookies(self, user, cookies):
        with HEADER_FILE.open('w') as f:
            json.dump(cookies, f, indent=2)



import logging
from urllib.parse import urlparse, parse_qs
from typing import List, Optional

from feeluown.excs import NoUserLoggedIn
from feeluown.library import (
    AbstractProvider, ProviderV2, ProviderFlags as Pf,
    SongModel, VideoModel, BriefVideoModel, BriefUserModel, ModelType,
    BriefPlaylistModel, BriefArtistModel, PlaylistModel, ModelNotFound,
)
from feeluown.media import Quality, Media, VideoAudioManifest, MediaType
from feeluown.library import SearchType, SimpleSearchResult
from feeluown.library.model_protocol import BriefSongProtocol

from fuo_ytmusic.consts import HEADER_FILE
from fuo_ytmusic.models import Categories, YtBriefUserModel, YtmusicWatchPlaylistSong
from fuo_ytmusic.service import YtmusicService, YtmusicType, YtmusicPrivacyStatus

logger = logging.getLogger(__name__)


class YtmusicProvider(AbstractProvider, ProviderV2):

    def __init__(self):
        super(YtmusicProvider, self).__init__()
        self.service: YtmusicService = YtmusicService()
        self._user = None
        self._http_proxy = ''

    def setup_http_proxy(self, http_proxy):
        self._http_proxy = http_proxy
        self.service.setup_http_proxy(http_proxy)

    def setup_http_timeout(self, timeout):
        self.service.setup_timeout(timeout)

    # noinspection PyPep8Naming
    class meta:
        identifier = 'ytmusic'
        name = 'YouTube Music'

    @property
    def identifier(self):
        return self.meta.identifier

    @property
    def name(self):
        return self.meta.name

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, user):
        self.service.reload()
        self._user = user

    def use_model_v2(self, mtype):
        return mtype in (
            ModelType.song,
            ModelType.album,
            ModelType.artist,
            ModelType.playlist,
        )

    def library_upload_songs(self):
        songs = self.service.library_upload_songs(100)
        return [song.model() for song in songs]

    def library_upload_artists(self):
        artists = self.service.library_upload_artists(100)
        return [artist.model() for artist in artists]

    def library_upload_albums(self):
        albums = self.service.library_upload_albums(100)
        return [album.model() for album in albums]

    def library_songs(self):
        songs = self.service.library_songs(100)
        return [song.v2_model() for song in songs]

    def library_albums(self):
        albums = self.service.library_albums(100)
        return [album.v2_brief_model() for album in albums]

    def library_artists(self) -> List[BriefArtistModel]:
        artists = self.service.library_subscription_artists(100)
        return [artist.v2_brief_model() for artist in artists]

    def library_playlists(self) -> List[BriefPlaylistModel]:
        playlists = self.service.library_playlists(100)
        return [playlist.v2_brief_model() for playlist in playlists]

    def create_playlist(self, title: str, description: str, privacy_status: YtmusicPrivacyStatus,
                        video_ids: List[str] = None, source_playlist: str = None) -> bool:
        return self.service.create_playlist(title, description, privacy_status, video_ids, source_playlist)

    def playlist_info(self, identifier) -> PlaylistModel:
        return self.service.playlist_info(identifier, limit=0).v2_model()

    def add_playlist_item(self, identifier, song_id) -> bool:
        result = self.service.add_playlist_items(identifier, [song_id])
        return result.status == 'STATUS_SUCCEEDED'

    def remove_playlist_item(self, identifier, song_id, set_song_id) -> bool:
        return self.service.remove_playlist_items(identifier, [{
            'videoId': song_id,
            'setVideoId': set_song_id,
        }]) == 'STATUS_SUCCEEDED'

    def category_playlists(self, params):
        playlists = self.service.category_playlists(params)
        return [p.model() for p in playlists]

    def categories(self) -> List[Categories]:
        return self.service.categories()

    def user_from_cookie(self, cookies: dict):
        return YtBriefUserModel(
            identifier='', source=self.meta.identifier, name='Me',
            cookies=cookies)

    def has_current_user(self) -> bool:
        return self._user is not None

    def get_current_user(self):
        if not self._user:
            raise NoUserLoggedIn
        return self._user

    def user_get(self, identifier):
        if identifier is None:
            return None
        if identifier == '':
            return BriefUserModel(identifier='', source=self.meta.identifier, name='Me')
        user = self.service.user_info(identifier)
        return BriefUserModel(identifier=identifier, source=self.meta.identifier, name=user.name)

    def search(self, keyword, type_, *args, **kwargs):
        type_ = SearchType.parse(type_)
        ytmusic_type = YtmusicType.parse(type_)
        results = self.service.search(keyword, ytmusic_type)
        model = SimpleSearchResult(q=keyword)
        if results:
            try:
                results[0].v2_model
            except AttributeError:
                models = [r.v2_brief_model() for r in results]
            else:
                # Try to use SongModel instead of BriefSongModel,
                # because there is no way to implement song_get.
                models = [r.v2_model() for r in results]
        else:
            models = []
        setattr(model, ytmusic_type.value, models)
        return model

    def song_list_quality(self, song) -> List[Quality.Audio]:
        return []

        id_ = song.identifier
        song_ = self.service.song_info(id_)
        return song_.list_formats() if song_ is not None else []

    def song_get_media(self, song: SongModel, quality: Quality.Audio) -> Optional[Media]:
        return None

        media = self._get_media(song, quality)
        if media is None:
            return media
        url = media.url
        # 推断(cosven): service.song_info 接口返回的 url 里面会记录请求时的 IP，
        # 如果后面真正访问 url 时，如果自己的 IP 已经变了（比如自己的代理 IP 变了），
        # 那么会碰到 403 错误。
        #
        # 注：你或许会想，把 url 中的 IP 改变当前的 public IP，是不是就行了？
        #    这其实也是不行的，因为整个 url 是已经有摘要信息的，它要和摘要匹配。
        if self.service.check_stream_url(url):
            return media
        parse_result = urlparse(url)
        kvs = parse_qs(parse_result.query)
        ips = kvs.get('ip', [])
        ip = ips[0] if ips else ''
        logger.info(
            f"url for video({song.identifier}) is invalid now, will retry! "
            f"maybe your public IP is changed (expected ip: {ip} ), (url: {url} )"
        )
        return self._get_media(song, quality)

    def _get_media(self, song, quality: Quality.Audio):
        song_info = self.service.song_info(song.identifier)
        format_code, bitrate, format_str = song_info.get_media(quality)
        url = self.service.stream_url(song_info, song.identifier, format_code)
        if url is not None:
            if 'video/mp4' in format_str:
                format_ = 'mp4'
            elif 'audio/mp4' in format_str:
                format_ = 'm4a'
            else:
                format_ = ''
            return Media(url, type_=MediaType.audio, bitrate=bitrate,
                         format=format_, http_proxy=self._http_proxy)
        return None

    def song_get_web_url(self, song) -> str:
        return f'https://music.youtube.com/watch?v={song.identifier}'

    def song_get(self, identifier):
        # ytmusicapi has not api to get song detail.
        # hack(cosven): we use get_watch_playlist to try to get song detail.
        # It works for song like '如愿-王菲'.
        result = self.service.api.get_watch_playlist(identifier)
        songs = [YtmusicWatchPlaylistSong(**track).v2_model()
                 for track in result['tracks']]
        for song in songs:
            if song.identifier == identifier:
                return song
        # I think this branch should not be reached (in most cases).
        return ModelNotFound(f'song:{identifier} not found')

    def song_list_similar(self, song):
        result = self.service.api.get_watch_playlist(song.identifier)
        songs = [YtmusicWatchPlaylistSong(**track).v2_model()
                 for track in result['tracks']
                 if track['videoId'] != song.identifier]
        return songs

    def album_get(self, identifier):
        album_info = self.service.album_info(identifier)
        return album_info.v2_model_with_identifier(identifier)

    def artist_get(self, identifier):
        return self.service.artist_info(identifier).v2_model(identifier)

    def playlist_get(self, identifier):
        return self.service.playlist_info(identifier).v2_model()

    def playlist_create_songs_rd(self, playlist):
        playlist_info = self.service.playlist_info(playlist.identifier)
        return playlist_info.reader(self)

    def playlist_add_song(self, playlist, song):
        if playlist.identifier == 'LM':
            return False
        return self.add_playlist_item(playlist.identifier, song.identifier)

    # playlist.set_id_map is not currently maintained.
    #
    # def playlist_remove_song(self, playlist, song):
    #     song_id = song.identifier
    #     if playlist.identifier == 'LM':
    #         return False
    #     set_id = playlist.set_id_map.get(song_id)
    #     if set_id is None:
    #         return False
    #     return self.remove_playlist_item(self.identifier, song_id, set_id)

    def artist_create_songs_rd(self, artist):
        artist_info = self.service.artist_info(artist.identifier)
        if artist_info.songs.browseId is None:
            # results may also be none.
            # for example: channelId=UCGSXa1Ve1FswQxtwarGi-Vg
            return [song.v2_model() for song in artist_info.songs.results or []]
        playlist_info = self.service.playlist_info(artist_info.songs.browseId)
        return playlist_info.reader(self)

    def artist_create_albums_rd(self, artist):
        artist_info = self.service.artist_info(artist.identifier)
        # Sometimes, the artist only has few albums, read them from results.
        if artist_info.albums.browseId is None:
            albums = artist_info.albums.results
        else:
            albums = self.service.artist_albums(artist_info.albums.browseId,
                                                artist_info.albums.params)
        return [album.v2_brief_model() for album in albums]

    def deprecated_song_get_lyric(self, song):
        # 歌词获取报错的 workaround
        if self._app is None:
            return None
        try:
            from fuo_netease.provider import NeteaseProvider
            provider: NeteaseProvider = self._app.library.get('netease')
            if provider is None:
                return None
            result = provider.search(f'{song.title} {song.artists_name}', type_=SearchType.so)
            songs = result.songs
            if len(songs) < 1:
                return None
            return provider.song_get_lyric(songs[0])
        except ModuleNotFoundError:
            return None
        except:
            raise

    def video_list_quality(self, video) -> List[Quality.Video]:
        return []

        id_ = video.identifier
        song_ = self.service.song_info(id_)
        return song_.list_video_formats() if song_ is not None else []

    def video_get(self, identifier):
        song = self.song_get(identifier)
        return VideoModel(identifier=song.identifier, source=song.source, title=song.title,
                          artists=song.artists, duration=song.duration, cover=song.pic_url)

    def video_get_web_url(self, video) -> str:
        return f'https://youtube.com/watch?v={video.identifier}'

    def video_get_media(self, video, quality) -> Optional[Media]:
        return None

        song_info = self.service.song_info(video.identifier)
        format_code = song_info.get_mv(quality)
        audio_formats = song_info.list_formats()
        audio_code, _, __ = song_info.get_media(audio_formats[0])
        url = self.service.stream_url(song_info, video.identifier, format_code)
        audio_url = self.service.stream_url(song_info, video.identifier, audio_code)
        if url is None or audio_url is None:
            return None
        return Media(VideoAudioManifest(url, audio_url), http_proxy=self._http_proxy)

    def song_get_mv(self, song: BriefSongProtocol) -> BriefVideoModel:
        return BriefVideoModel(identifier=song.identifier, source=song.source, title=song.title,
                               artists_name=song.artists_name, duration_ms=song.duration_ms)

    def upload_song(self, path: str) -> bool:
        return self.service.upload_song(path) == 'STATUS_SUCCEEDED'

    def delete_uploaded_song(self, entity_id: str) -> bool:
        return self.service.delete_upload_song(entity_id) == 'STATUS_SUCCEEDED'


provider = YtmusicProvider()



import asyncio
import os
from pathlib import Path

from PyQt5.QtCore import Qt, QUrl, QObject, pyqtProperty, pyqtSlot, pyqtSignal
from PyQt5.QtQuick import QQuickView
from PyQt5.QtWidgets import QWidget

from feeluown.library import BriefPlaylistModel
from feeluown.gui.base_renderer import TabBarRendererMixin
from fuo_ytmusic import YtmusicProvider


class ExploreBackend(QObject):
    categoriesLoaded = pyqtSignal('QVariantList')
    playlistsLoaded = pyqtSignal('QVariantList')

    def __init__(self, provider: YtmusicProvider, app):
        super().__init__()
        self._provider = provider
        self._app = app

    async def categories(self):
        loop = asyncio.get_event_loop()
        categories = await loop.run_in_executor(None, self._provider.categories)
        result = [{'key': category.key,
                   'value': [{'title': playlist.title, 'params': playlist.params} for playlist in category.value]} for
                  category in categories]
        self.categoriesLoaded.emit(result)
        return categories

    async def playlists(self, params: str):
        loop = asyncio.get_event_loop()
        playlists = await loop.run_in_executor(None, self._provider.service.category_playlists, params)
        result = [{'id': p.playlistId, 'name': p.title, 'cover': p.thumbnail} for p in playlists]
        self.playlistsLoaded.emit(result)

    @pyqtSlot()
    def load_categories(self):
        asyncio.ensure_future(self.categories())

    @pyqtSlot(str)
    def load_playlists(self, params: str):
        asyncio.ensure_future(self.playlists(params))

    @pyqtSlot(str, str, str)
    def goto_playlist(self, playlist_id: str, name: str, cover: str):
        model = BriefPlaylistModel(identifier=playlist_id, source='ytmusic', name=name)
        self._app.browser.goto(model=model)

    @pyqtProperty(bool, constant=True)
    def is_dark(self) -> bool:
        return self._app.theme_mgr.theme == 'dark'


async def render(req, **_):
    app = req.ctx['app']
    provider = app.library.get('ytmusic')
    backend = ExploreBackend(provider, app)
    categories = await backend.categories()
    tab_index: int = int(req.query.get('flow_index', 0))
    tabs = [(c.key, i) for i, c in enumerate(categories)]
    renderer = ExploreRenderer(app, tab_index, tabs, backend)
    await renderer.render()
    backend._x_renderer = renderer  # HACK: prevent renderer objec to be freed.


class ExploreRenderer(TabBarRendererMixin):

    def __init__(self, app, tab_index, tabs, backend):
        self._app = app
        self._backend = backend
        self.tab_index = tab_index
        self.tabs = tabs

    async def render(self):
        self.render_tab_bar()

        os.environ['QT_QUICK_CONTROLS_STYLE'] = 'Material'
        view = QQuickView()
        view.rootContext().setContextProperty('explore_backend', self._backend)
        view.rootContext().setContextProperty('flow_index', self.get_flow_index(self.tab_index))
        container = QWidget.createWindowContainer(view, self._app.ui.right_panel)
        container.setFocusPolicy(Qt.TabFocus)
        view.setResizeMode(QQuickView.SizeRootObjectToView)
        view.setSource(QUrl.fromLocalFile((Path(__file__).parent / 'qml' / 'page_explore.qml').as_posix()))
        self._app.ui.right_panel.set_body(container)

    def render_by_tab_index(self, tab_index):
        self._app.browser.goto(page='/providers/ytmusic/explore',
                               query={'flow_index': self.get_flow_index(tab_index)})

    def get_flow_index(self, tab_index):
        return self.tabs[tab_index][1]



import logging
import ntpath
import os
import re
import sys
import threading
from datetime import timedelta
from enum import Enum
from functools import partial
from typing import Optional, Union, List
from urllib.parse import unquote

import requests
from ytmusicapi import YTMusic as YTMusicBase
from ytmusicapi.ytmusic import OAuthCredentials
from cachetools.func import ttl_cache
from requests import Response
from feeluown.library import SearchType

from fuo_ytmusic.cipher import Cipher
from fuo_ytmusic.consts import HEADER_FILE
from fuo_ytmusic.helpers import Singleton
from fuo_ytmusic.models import (
    YtmusicSearchSong,
    YtmusicSearchAlbum,
    YtmusicSearchArtist,
    YtmusicSearchVideo,
    YtmusicSearchPlaylist,
    YtmusicSearchBase,
    YtmusicDispatcher,
    ArtistInfo,
    UserInfo,
    AlbumInfo,
    SongInfo,
    Categories,
    PlaylistNestedResult,
    TopCharts,
    YtmusicLibrarySong,
    YtmusicLibraryArtist,
    PlaylistInfo,
    YtmusicHistorySong,
    PlaylistAddItemResponse,
)


CACHE_TTL = timedelta(minutes=10).seconds
CACHE_SIZE = 1
GLOBAL_LIMIT = 20

logger = logging.getLogger(__name__)


class YtmusicType(Enum):
    so = "songs"
    vi = "videos"
    ar = "artists"
    al = "albums"
    pl = "playlists"

    # noinspection PyTypeChecker
    @classmethod
    def parse(cls, type_: SearchType) -> "YtmusicType":
        return cls._value2member_map_.get(type_.value + "s")


class YtmusicPrivacyStatus(Enum):
    PUBLIC = "PUBLIC"
    PRIVATE = "PRIVATE"
    UNLISTED = "UNLISTED"

    # noinspection PyTypeChecker
    @classmethod
    def parse(cls, type_: str) -> "YtmusicPrivacyStatus":
        return cls._value2member_map_.get(type_)


class YtmusicScope(Enum):
    li = "library"
    up = "uploads"


class YTMusic(YTMusicBase):
    class IterableToFileAdapter(object):
        def __init__(self, iterable):
            self.iterator = iter(iterable)
            self.length = len(iterable)

        def read(self, size=-1):  # TBD: add buffer for `len(data) > size` case
            return next(self.iterator, b"")

        def __len__(self):
            return self.length

    class ChunckedUpload(object):
        def __init__(self, filename, chunksize=1 << 6):
            self.filename = filename
            self.chunksize = chunksize
            self.totalsize = os.path.getsize(filename)
            self.readsofar = 0

        def __iter__(self):
            with open(self.filename, "rb") as file:
                while True:
                    data = file.read(self.chunksize)
                    if not data:
                        sys.stdout.write("\n")
                        break
                    self.readsofar += len(data)
                    percent = self.readsofar * 1e2 / self.totalsize
                    sys.stdout.write(
                        "\rUploading: {percent:3.0f}%".format(percent=percent)
                    )
                    yield data

        def __len__(self):
            return self.totalsize

    def upload_song_progress(self, filepath: str) -> Union[str, requests.Response]:
        """
        Uploads a song to YouTube Music

        :param filepath: Path to the music file (mp3, m4a, wma, flac or ogg)
        :return: Status String or full response
        """
        self._check_auth()
        if not os.path.isfile(filepath):
            raise Exception("The provided file does not exist.")

        supported_filetypes = ["mp3", "m4a", "wma", "flac", "ogg"]
        if os.path.splitext(filepath)[1][1:] not in supported_filetypes:
            raise Exception(
                "The provided file type is not supported by YouTube Music. Supported file types are "
                + ", ".join(supported_filetypes)
            )

        headers = self.headers.copy()
        upload_url = "https://upload.youtube.com/upload/usermusic/http?authuser=0"
        filesize = os.path.getsize(filepath)
        body = ("filename=" + ntpath.basename(filepath)).encode("utf-8")
        headers.pop("content-encoding", None)
        headers["content-type"] = "application/x-www-form-urlencoded;charset=utf-8"
        headers["X-Goog-Upload-Command"] = "start"
        headers["X-Goog-Upload-Header-Content-Length"] = str(filesize)
        headers["X-Goog-Upload-Protocol"] = "resumable"
        response = requests.post(
            upload_url, data=body, headers=headers, proxies=self.proxies
        )
        headers["X-Goog-Upload-Command"] = "upload, finalize"
        headers["X-Goog-Upload-Offset"] = "0"
        upload_url = response.headers["X-Goog-Upload-URL"]
        response = requests.post(
            upload_url,
            data=YTMusic.IterableToFileAdapter(YTMusic.ChunckedUpload(filepath)),
            headers=headers,
            proxies=self.proxies,
        )
        if response.status_code == 200:
            return "STATUS_SUCCEEDED"
        else:
            return response


class YtmusicService(metaclass=Singleton):
    def __init__(self):
        self._session = requests.Session()
        self._api: Optional[YTMusic] = None
        self._session.hooks["response"].append(self._do_logging)

        self._cipher = None
        self._signature_timestamp = 0
        self._cipher_lock = threading.Lock()
        self._api_lock = threading.Lock()

    @staticmethod
    def _do_logging(r: Response, *_, **__):
        logger.debug(
            f"[ytmusic] Requesting: [{r.request.method.upper()}] {r.url}; "
            f"Response: [{r.status_code}] {len(r.content)} bytes."
        )

    @property
    def api(self) -> YTMusic:
        if self._api is None:
            with self._api_lock:
                if self._api is None:
                    self._initialize_by_headerfile()
        return self._api

    def _log_thread(self):
        return f"Thread({threading.get_ident()})"

    def get_cipher(self):
        if self._cipher is None:
            logger.info(f"{self._log_thread()} try to get cipher...")
            with self._cipher_lock:
                if self._cipher is None:
                    js_url = self.api.get_basejs_url()
                    js = self._session.get(js_url).text
                    match = re.search(r"signatureTimestamp[:=](\d+)", js)
                    assert match is not None, "Unable to identify the signatureTimestamp."
                    self._signature_timestamp = int(match.group(1))
                    self._cipher = Cipher(js)
                    logger.info(f"{self._log_thread()} got cipher")
                else:
                    logger.info(f"{self._log_thread()} cipher already exists")
        return self._cipher

    def reset_cipher(self):
        with self._cipher_lock:
            # I don't know if cipher has some relation with signature timestamp.
            self._cipher = None
            self._signature_timestamp = 0

    def get_signature_timestamp(self):
        # This method works along with pytube cipher, which is used to get
        # a playable media url. However, pytube cipher became invalid since 2024-06.
        # The url processed by pytube cipher still returns 403.
        # It seems others also met this problem: https://github.com/pytube/pytube/issues/1943
        #
        # So return 0 directly to avoid get_cipher, get_cipher costs too much time.
        return 0

        if self._signature_timestamp == 0:
            logger.info(f"{self._log_thread()} signature timestamp is 0, try to refresh.")
            self.get_cipher()
        assert self._signature_timestamp != 0, "signature timestamp should not be 0 now."
        return self._signature_timestamp

    def _initialize_by_headerfile(self):
        options = dict(
            requests_session=self._session,
            language="zh_CN",
            oauth_credentials=OAuthCredentials(session=self._session),
        )
        if HEADER_FILE.exists():
            logger.info("Initializing ytmusic api with headerfile.")
            self._api = YTMusic(str(HEADER_FILE), **options)
        else:
            logger.info("Initializing ytmusic api with no headerfile.")
            # Actually, YTMusic does not work if no auth file is provided.
            self._api = YTMusic(**options)
        threading.Thread(target=self.get_signature_timestamp).start()

    def reload(self):
        self._initialize_by_headerfile()

    def setup_http_proxy(self, http_proxy):
        self._session.proxies = {
            "http": http_proxy,
            "https": http_proxy,
        }

    def setup_timeout(self, timeout):
        if isinstance(self._session.request, partial):
            request = self._session.request.func
        else:
            request = self._session.request
        self._session.request = partial(request, timeout=timeout)

    def search(
        self,
        keywords: str,
        t: Optional[YtmusicType],
        scope: YtmusicScope = None,
        page_size: int = GLOBAL_LIMIT,
    ) -> List[
        Union[
            YtmusicSearchSong,
            YtmusicSearchAlbum,
            YtmusicSearchArtist,
            YtmusicSearchVideo,
            YtmusicSearchPlaylist,
            YtmusicSearchBase,
        ]
    ]:
        response = self.api.search(
            keywords,
            None if t is None else t.value,
            None if scope is None else scope.value,
            page_size,
        )
        return [YtmusicDispatcher.search_result_dispatcher(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def artist_info(self, channel_id: str) -> ArtistInfo:
        data = self.api.get_artist(channel_id)
        return ArtistInfo(**data)

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def artist_albums(self, channel_id: str, params: str) -> List[YtmusicSearchAlbum]:
        response = self.api.get_artist_albums(channel_id, params)
        return [YtmusicSearchAlbum(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def user_info(self, channel_id: str) -> UserInfo:
        return UserInfo(**self.api.get_user(channel_id))

    def user_playlists(self, channel_id: str, params: str):
        return self.api.get_user_playlists(channel_id, params)

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def album_info(self, browse_id: str) -> AlbumInfo:
        data = self.api.get_album(browse_id)
        return AlbumInfo(**data)

    def song_info(self, video_id: str) -> SongInfo:
        return SongInfo(**self.api.get_song(video_id, self.get_signature_timestamp()))

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def categories(self) -> List[Categories]:
        return [
            Categories(key=k, value=v)
            for k, v in self.api.get_mood_categories().items()
        ]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def category_playlists(self, params: str) -> List[PlaylistNestedResult]:
        response = self.api.get_mood_playlists(params)
        return [PlaylistNestedResult(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def get_charts(self, country: str = "ZZ") -> TopCharts:
        # temp workaround for ytmusicapi#236
        # sees: https://github.com/sigma67/ytmusicapi/issues/236
        auth = self.api.auth
        self.api.auth = None
        response = self.api.get_charts(country)
        self.api.auth = auth
        return TopCharts(**response)

    def library_playlists(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[PlaylistNestedResult]:
        response = self.api.get_library_playlists(limit)
        return [PlaylistNestedResult(**data) for data in response]

    def library_songs(self, limit: int = GLOBAL_LIMIT) -> List[YtmusicLibrarySong]:
        response = self.api.get_library_songs(limit)
        return [YtmusicLibrarySong(**data) for data in response]

    def library_albums(self, limit: int = GLOBAL_LIMIT) -> List[YtmusicSearchAlbum]:
        response = self.api.get_library_albums(limit)
        return [YtmusicSearchAlbum(**data) for data in response]

    def library_artists(self, limit: int = GLOBAL_LIMIT) -> List[YtmusicLibraryArtist]:
        response = self.api.get_library_artists(limit)
        return [YtmusicLibraryArtist(**data) for data in response]

    def library_subscription_artists(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicLibraryArtist]:
        response = self.api.get_library_subscriptions(limit)
        return [YtmusicLibraryArtist(**data) for data in response]

    @ttl_cache(maxsize=CACHE_SIZE, ttl=CACHE_TTL)
    def playlist_info(
        self, playlist_id: str, limit: int = GLOBAL_LIMIT
    ) -> PlaylistInfo:
        return PlaylistInfo(**self.api.get_playlist(playlist_id, limit))

    def liked_songs(self, limit: int = GLOBAL_LIMIT) -> PlaylistInfo:
        return PlaylistInfo(**self.api.get_liked_songs(limit))

    def history(self) -> List[YtmusicHistorySong]:
        response = self.api.get_history()
        return [YtmusicHistorySong(**data) for data in response]

    def create_playlist(
        self,
        title: str,
        description: str,
        privacy_status: YtmusicPrivacyStatus,
        video_ids: List[str] = None,
        source_playlist: str = None,
    ) -> bool:
        response = self.api.create_playlist(
            title, description, privacy_status.value, video_ids, source_playlist
        )
        if not isinstance(response, str):
            return False
        return True

    def add_playlist_items(
        self,
        playlist_id: str,
        video_ids: List[str] = None,
        source_playlist_id: str = None,
    ) -> PlaylistAddItemResponse:
        return PlaylistAddItemResponse(
            **self.api.add_playlist_items(playlist_id, video_ids, source_playlist_id)
        )

    def remove_playlist_items(
        self, playlist_id: str, video_ids: List[dict]
    ) -> Optional[str]:
        # STATUS_SUCCEEDED STATUS_FAILED
        return self.api.remove_playlist_items(playlist_id, video_ids)

    def library_upload_songs(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicLibrarySong]:
        response = self.api.get_library_upload_songs(limit)
        return [YtmusicLibrarySong(**data) for data in response]

    def library_upload_artists(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicLibraryArtist]:
        response = self.api.get_library_upload_artists(limit)
        return [YtmusicLibraryArtist(**data) for data in response]

    def library_upload_albums(
        self, limit: int = GLOBAL_LIMIT
    ) -> List[YtmusicSearchAlbum]:
        response = self.api.get_library_upload_albums(limit)
        return [YtmusicSearchAlbum(**data) for data in response]

    def upload_song(self, file: str) -> str:
        # STATUS_SUCCEEDED
        return self.api.upload_song_progress(file)

    def delete_upload_song(self, entity_id: str) -> str:
        # STATUS_SUCCEEDED
        return self.api.delete_upload_entity(entity_id)

    def stream_url(self, song_info: SongInfo, video_id: str, format_code: int) -> Optional[str]:
        formats = song_info.streamingData.adaptiveFormats
        for f in formats:
            if int(f.itag) == format_code:
                return self._get_stream_url(f, video_id)
        return None

    def check_stream_url(self, url):
        resp = self._session.head(url)
        return resp.status_code != 403

    def _get_stream_url(self, f: SongInfo.StreamingData.Format, video_id: str) -> Optional[str]:
        if f.url is not None and f.url != "":
            return f.url
        sig_ch = f.signatureCipher
        sig_ex = sig_ch.split("&")
        res = dict({"s": "", "url": ""})
        for sig in sig_ex:
            for key in res:
                if sig.find(key + "=") >= 0:
                    res[key] = unquote(sig[len(key + "=") :])
        signature = self.get_cipher().get_signature(ciphered_signature=res["s"])
        _url = res["url"] + "&sig=" + signature
        return _url


if __name__ == "__main__":
    # noinspection PyUnresolvedReferences

    service = YtmusicService()
    # print(service.upload_song('/home/bruce/Music/阿梓 - 呼吸决定.mp3'))
    print(service.api.delete_upload_entity("t_po_CJT896eTt_a5swEQwrmdzf7_____AQ"))



from typing import Optional, Union, List, Tuple

# noinspection PyProtectedMember
from pydantic.v1.fields import Field
from pydantic.v1 import BaseModel as PydanticBaseModel
from pydantic.v1.main import ModelMetaclass

from feeluown.utils.reader import SequentialReader
from feeluown.media import Quality
from feeluown.library import (
    VideoModel, ModelState, BriefArtistModel, BriefUserModel,
    AlbumModel as AlbumModelV2, BriefSongModel, SongModel as SongModelV2,
    BriefAlbumModel, ArtistModel as ArtistModelV2, PlaylistModel,
    BriefPlaylistModel,
)
from feeluown.library import AlbumType

from fuo_ytmusic.timeparse import timeparse


class AllowOptional(ModelMetaclass):
    def __new__(mcs, name, bases, namespaces, **kwargs):
        annotations = namespaces.get('__annotations__', {})
        for base in bases:
            if hasattr(base, '__annotations__'):
                annotations = {**annotations, **base.__annotations__}
        for field in annotations:
            if not field.startswith('__'):
                if hasattr(annotations[field], '__origin__') and annotations[field].__origin__ is Union:
                    continue
                annotations[field] = Optional[annotations[field]]
        namespaces['__annotations__'] = annotations
        return super().__new__(mcs, name, bases, namespaces, **kwargs)


class BaseModel(PydanticBaseModel, metaclass=AllowOptional):
    @property
    def source(self):
        return 'ytmusic'


class SearchNestedArtist(BaseModel):
    id: str
    name: str

    def v2_brief_model(self) -> BriefArtistModel:
        return BriefArtistModel(identifier=self.id or '', source=self.source, name=self.name or '')


class YtmusicArtistsMixin:
    artists: List[SearchNestedArtist]  # 歌手信息

    def v2_brief_artist_models(self) -> List[BriefArtistModel]:
        return [artist.v2_brief_model() for artist in (self.artists or [])
                if artist.id is not None]

    @property
    def artists_name(self):
        if self.artists is None:
            return ''
        return ' / '.join([a.name for a in self.artists])


class SearchNestedThumbnail(BaseModel):
    url: str  # 图片地址
    width: int
    height: int


class YtmusicCoverMixin:
    thumbnails: List[SearchNestedThumbnail]  # 封面信息

    @property
    def cover(self) -> Optional[str]:
        if self.thumbnails is not None and len(self.thumbnails) > 0:
            return self.thumbnails[-1].url
        return None

    @property
    def thumbnail(self) -> Optional[str]:
        if self.thumbnails is not None and len(self.thumbnails) > 0:
            return self.thumbnails[0].url
        return None


class YtmusicDurationMixin:
    duration: str  # 歌曲时长 eg.3:50

    @property
    def duration_ms(self) -> int:
        if self.duration is None:
            return 0
        return int(timeparse(self.duration) * 1000)


class YtmusicAlbumSong(BaseModel, YtmusicArtistsMixin, YtmusicDurationMixin):
    title: str
    album: str
    videoId: str

    def v2_brief_model(self) -> BriefSongModel:
        return BriefSongModel(
            identifier=self.videoId,
            source=self.source,
            title=self.title,
            artists_name=self.artists_name,
            album_name=self.album,
            duration_ms=self.duration
        )

    def v2_model_with_brief_album(self, album: BriefAlbumModel) -> SongModelV2:
        return SongModelV2(
            identifier=self.videoId,
            source=self.source,
            title=self.title,
            album=album,
            artists=self.v2_brief_artist_models(),
            duration=self.duration_ms,
        )


class YtmusicSearchBase(BaseModel):
    category: str
    resultType: str


class YtmusicSearchSong(YtmusicSearchBase, YtmusicCoverMixin, YtmusicArtistsMixin, YtmusicDurationMixin):
    class Album(BaseModel):
        id: str
        name: str

        def model(self) -> BriefAlbumModel:
            album = BriefAlbumModel(identifier=self.id or '', source=self.source, name=self.name)
            if self.id is None:
                album.state = ModelState.not_exists
            return album

    title: str  # 歌名
    album: Album  # 专辑信息
    feedbackTokens: dict
    videoId: str  # 歌曲ID
    isAvailable: bool
    isExplicit: bool

    def v2_brief_model(self) -> BriefSongModel:
        song = BriefSongModel(
            identifier=self.videoId or '',
            source=self.source,
            title=self.title,
            artists_name=self.artists_name,
            album_name=self.album.name if self.album else '',
            duration_ms=self.duration
        )
        if not song.identifier:
            song.state = ModelState.not_exists
        return song

    def v2_model(self) -> SongModelV2:
        if self.album:
            album = self.album.model()
        else:
            album = None
        song = SongModelV2(
            identifier=self.videoId or '',
            source=self.source,
            title=self.title,
            artists=self.v2_brief_artist_models(),
            album=album,
            duration=self.duration_ms,
            pic_url=self.cover or '',
        )
        if not song.identifier:
            song.state = ModelState.not_exists
        return song


class YtmusicWatchPlaylistSong(YtmusicSearchSong):
    year: str  # This field exists in get_watch_playlist API.

    def v2_model(self) -> SongModelV2:
        song = super().v2_model()
        song.date = self.year or ''
        return song


class YtmusicLibrarySong(YtmusicSearchSong):
    likeStatus: str  # LIKE
    setVideoId: str
    entityId: str


class YtmusicHistorySong(YtmusicLibrarySong):
    played: str  # 上次播放 eg November 2021


class YtmusicSearchAlbum(YtmusicSearchBase, YtmusicCoverMixin, YtmusicArtistsMixin):
    title: str  # 专辑名
    type: str  # 专辑类型
    year: Optional[Union[int, str]]  # 年
    browseId: str  # 查询ID
    isExplicit: bool

    @property
    def album_type(self) -> AlbumType:
        if self.type == 'Single':
            return AlbumType.single
        return AlbumType.standard

    def v2_brief_model(self) -> BriefAlbumModel:
        return BriefAlbumModel(
            identifier=self.browseId,
            source=self.source,
            name=self.title,
            artists_name=self.artists_name,
        )


class YtmusicSearchArtist(YtmusicSearchBase, YtmusicCoverMixin):
    artist: str  # 歌手名
    shuffleId: str
    radioId: str
    browseId: str  # 查询ID

    def v2_brief_model(self) -> BriefArtistModel:
        return BriefArtistModel(identifier=self.browseId, source=self.source, name=self.artist)


class YtmusicLibraryArtist(YtmusicSearchArtist):
    subscribers: str  # 歌曲数量


class YtmusicSearchPlaylist(YtmusicSearchBase, YtmusicCoverMixin):
    title: str  # 歌单名
    itemCount: Optional[int] = None  # 歌曲数量
    author: str  # 歌单作者
    browseId: str  # 查询ID

    def v2_brief_model(self) -> BriefPlaylistModel:
        return BriefPlaylistModel(
            identifier=self.browseId,
            source=self.source,
            name=self.title,
            creator_name=self.author or '',
        )


class YtmusicSearchVideo(YtmusicSearchBase, YtmusicCoverMixin, YtmusicArtistsMixin, YtmusicDurationMixin):
    title: str  # 视频标题
    views: str  # 播放量 eg:13K
    videoId: str  # 视频ID
    playlistId: str

    def v2_model(self) -> VideoModel:
        return VideoModel(
            identifier=self.videoId,
            source=self.source,
            title=self.title,
            cover=self.cover,
            artists=self.v2_brief_artist_models(),
            duration=self.duration_ms
        )


class YtmusicDispatcher:
    RESULT_TYPE_MAP = {
        'video': YtmusicSearchVideo,
        'song': YtmusicSearchSong,
        'artist': YtmusicSearchArtist,
        'album': YtmusicSearchAlbum,
        'playlist': YtmusicSearchPlaylist,
    }

    @classmethod
    def search_result_dispatcher(cls, **data) \
            -> Union[YtmusicSearchBase, YtmusicSearchVideo, YtmusicSearchSong, YtmusicSearchArtist, YtmusicSearchAlbum,
                     YtmusicSearchPlaylist]:
        clazz = cls.RESULT_TYPE_MAP.get(data.get('resultType') or '')
        return clazz(**data) if clazz is not None else YtmusicSearchBase(**data)


class ArtistInfo(BaseModel):
    class Songs(BaseModel):
        browseId: str  # 查询ID
        results: List[YtmusicSearchSong]  # 歌曲列表（部分）

    class Albums(BaseModel):
        browseId: str  # 查询ID
        results: List[YtmusicSearchAlbum]  # 专辑列表（部分）
        params: str  # 不知道是啥

    class Videos(BaseModel):
        browseId: str  # 查询ID
        results: List[YtmusicSearchVideo]  # 视频列表（部分）

    class RelatedArtists(BaseModel):
        results: List[YtmusicSearchArtist]  # 相关歌手

    name: str  # 歌手名
    description: str  # 描述
    views: str  # 播放量 eg.177,865,792 views
    channelId: str
    shuffleId: str
    radioId: str
    subscribers: str  # 订阅量 eg.230K
    subscribed: bool  # 是否已订阅
    thumbnails: List[SearchNestedThumbnail]  # 封面信息
    songs: Songs
    albums: Albums  # 专辑
    singles: Albums  # 单曲专辑
    videos: Videos
    related: RelatedArtists

    def v2_model(self, identifier) -> ArtistModelV2:
        # Note that the channelId is different from the identifier.
        # Though the channelId also refers to the artist,
        # it's songs is a empty list.
        return ArtistModelV2(
            identifier=identifier,
            source=self.source,
            name=self.name,
            pic_url=(self.thumbnails[0].url if self.thumbnails else ''),
            aliases=[],
            hot_songs=[],
            description=self.description or '',
        )


class AlbumInfo(BaseModel, YtmusicArtistsMixin, YtmusicCoverMixin):
    title: str  # 专辑名
    type: str
    year: str
    trackCount: int
    duration: str  # eg.5 minutes, 14 seconds
    audioPlaylistId: str
    tracks: List[YtmusicAlbumSong]  # 专辑歌曲
    # ytmusicapi.get_album has this field. Not sure if other api has this field.
    description: str = ''

    def v2_model_with_identifier(self, identifier) -> AlbumModelV2:
        brief_album = BriefAlbumModel(
            identifier=identifier,
            source=self.source,
            name=self.title,
            artists_name=self.artists_name,
        )
        return AlbumModelV2(
            identifier=identifier,
            source=self.source,
            name=self.title,
            cover=self.cover,
            songs=[t.v2_model_with_brief_album(brief_album) for t in self.tracks],
            artists=self.v2_brief_artist_models(),
            description=self.description,
            released=self.year,
        )


class SongInfo(BaseModel):
    class VideoDetails(BaseModel):
        class Thumbnails(BaseModel, YtmusicCoverMixin):
            pass

        videoId: str
        title: str
        lengthSeconds: int
        channelId: str
        isOwnerViewing: bool
        isCrawlable: bool
        thumbnail: Thumbnails
        averageRating: float
        allowRatings: bool
        viewCount: int
        author: str
        isPrivate: bool
        musicVideoType: str
        isLiveContent: bool

    class StreamingData(BaseModel):
        class Format(BaseModel):
            itag: int
            url: str
            mimeType: str
            bitrate: int
            initRange: dict
            indexRange: dict
            lastModified: str
            contentLength: int
            audioQuality: str  # AUDIO_QUALITY_LOW AUDIO_QUALITY_MEDIUM
            audioSampleRate: int  # 48000
            quality: str
            signatureCipher: str

        expiresInSeconds: int
        formats: List[Format]
        adaptiveFormats: List[Format]

    videoDetails: VideoDetails
    streamingData: StreamingData

    def list_formats(self) -> List[Quality.Audio]:
        qualities = set()
        if self.streamingData is None:
            return []
        for format_ in self.streamingData.adaptiveFormats:
            if format_.audioQuality is None:
                continue
            if format_.audioQuality == 'AUDIO_QUALITY_LOW':
                qualities.add(Quality.Audio.lq)
            elif format_.audioQuality == 'AUDIO_QUALITY_MEDIUM':
                qualities.add(Quality.Audio.sq)
            elif format_.audioQuality == 'AUDIO_QUALITY_HIGH':
                qualities.add(Quality.Audio.hq)
        return list(qualities)

    def list_video_formats(self) -> List[Quality.Video]:
        qualities = set()
        for format_ in self.streamingData.adaptiveFormats:
            if format_.audioQuality is not None:
                continue
            if format_.quality == 'hd1080':
                qualities.add(Quality.Video.fhd)
            if format_.quality == 'hd720':
                qualities.add(Quality.Video.hd)
            if format_.quality == 'large':
                qualities.add(Quality.Video.sd)
            if format_.quality == 'medium':
                qualities.add(Quality.Video.ld)
        return list(qualities)

    def get_media(self, quality: Quality.Audio) -> Tuple[Optional[int], Optional[int], Optional[str]]:
        for format_ in self.streamingData.adaptiveFormats:
            if format_.audioQuality is None:
                continue
            if quality in (Quality.Audio.hq, Quality.Audio.shq) and format_.audioQuality == 'AUDIO_QUALITY_HIGH':
                return format_.itag, int(format_.bitrate / 1024), format_.mimeType
            if quality == Quality.Audio.sq and format_.audioQuality == 'AUDIO_QUALITY_MEDIUM':
                return format_.itag, int(format_.bitrate / 1024), format_.mimeType
            if quality == Quality.Audio.lq and format_.audioQuality == 'AUDIO_QUALITY_LOW':
                return format_.itag, int(format_.bitrate / 1024), format_.mimeType
        return None, None, None

    def get_mv(self, quality) -> Optional[int]:
        for format_ in self.streamingData.adaptiveFormats:
            if format_.audioQuality is not None:
                continue
            if quality == Quality.Video.fhd and format_.quality == 'hd1080':
                return format_.itag
            if quality == Quality.Video.hd and format_.quality == 'hd720':
                return format_.itag
            if quality == Quality.Video.sd and format_.quality == 'large':
                return format_.itag
            if quality == Quality.Video.ld and format_.quality == 'medium':
                return format_.itag
        return None


class PlaylistNestedResult(BaseModel, YtmusicCoverMixin):
    title: str
    playlistId: str

    def v2_brief_model(self) -> BriefPlaylistModel:
        return BriefPlaylistModel(
            identifier=self.playlistId,
            source=self.source,
            name=self.title,
            creator_name='',
        )


class UserInfo(BaseModel):
    class Playlists(BaseModel):
        browseId: str
        results: List[PlaylistNestedResult]
        params: str

    class Videos(BaseModel):
        browseId: str  # 查询ID
        results: List[YtmusicSearchVideo]  # 视频列表（部分）

    name: str
    playlists: Playlists
    videos: Videos


class Categories(BaseModel):
    class Category(BaseModel):
        title: str
        params: str

    key: str
    value: List[Category]


class TopCharts(BaseModel):
    class Countries(BaseModel):
        class Selected(BaseModel):
            text: str

        selected: Selected
        options: List[str]

    class Videos(BaseModel):
        playlist: str  # PlaylistID
        items: List[YtmusicSearchVideo]  # 视频列表（部分）

    class Artists(BaseModel):
        items: List[YtmusicSearchArtist]  # 歌手列表

    countries: Countries
    videos: Videos
    artists: Artists


class PlaylistInfo(BaseModel, YtmusicCoverMixin):
    class Author(BaseModel):
        id: str
        name: str

    id: str
    privacy: str  # PUBLIC
    title: str  # 歌单名
    description: Optional[str]
    author: Author
    year: int
    duration: str
    trackCount: int
    tracks: List[YtmusicLibrarySong]
    fetched_tracks: set = Field(default_factory=set)

    def v2_model(self):
        creator = None
        if self.author is not None and self.author.id is not None:
            creator = BriefUserModel(
                identifier=self.author.id,
                source=self.source,
                name=self.author.name or '',
            )
        return PlaylistModel(
            identifier=self.id,
            source=self.source,
            name=self.title,
            cover=self.cover,
            description=self.description or '',
            creator=creator
        )

    def reader(self, provider) -> SequentialReader:
        total_count = self.trackCount
        self.fetched_tracks = set()

        def g():
            counter = 0
            offset = 0
            per = 50
            while offset < total_count:
                end = min(offset + per, total_count)
                data: PlaylistInfo = provider.service.playlist_info(self.id, limit=end)
                tracks_data = data.tracks
                for track_data in tracks_data:
                    if track_data.videoId is not None and track_data.videoId in self.fetched_tracks:
                        continue
                    self.fetched_tracks.add(track_data.videoId)
                    counter += 1
                    yield track_data.v2_brief_model()
                if counter >= total_count:
                    break
                offset += per
            self.fetched_tracks.clear()

        return SequentialReader(g(), total_count)


class PlaylistAddItemResponse(BaseModel):
    class PlaylistEditResult(BaseModel):
        videoId: str
        setVideoId: str

    status: str  # STATUS_SUCCEEDED STATUS_FAILED
    playlistEditResults: List[PlaylistEditResult]


# FeelUOwn models

class YtmusicSongModel(SongModelV2):
    setVideoId: Optional[str]
    entityId: Optional[str]


class YtBriefUserModel(BriefUserModel):
    cookies: dict = {}



import asyncio
import logging
import threading
import time
from pathlib import Path
from typing import Optional, List

from PyQt5.QtCore import QObject, QUrl, pyqtSlot, QVariant, Qt, QModelIndex, QAbstractTableModel, pyqtProperty
from PyQt5.QtQuick import QQuickView
from PyQt5.QtWidgets import QFileDialog, QMessageBox
from feeluown.gui.widgets import TextButton
from feeluown.gui.widgets.songs import SongsTableView
from feeluown.utils import aio
from feeluown.gui.base_renderer import LibraryTabRendererMixin
from feeluown.gui.page_containers.table import Renderer
from feeluown.gui.widgets.tabbar import Tab

from fuo_ytmusic import YtmusicProvider

logger = logging.getLogger(__name__)


class UploadingSongListModel(QAbstractTableModel):
    def __init__(self, header: List[str], data: list):
        self._header = header
        self._data = data
        super(UploadingSongListModel, self).__init__()

    def update_state(self, index: int, state: bool):
        # noinspection PyUnresolvedReferences
        self.dataChanged.emit(self.index(index, 0), self.index(index, len(self._header) - 1))
        self._data[index][1] = '上传成功' if state else '上传失败'

    def insert_data(self, new_data: list):
        last_index = self.rowCount()
        self.beginInsertRows(QModelIndex(), last_index, last_index + len(new_data) - 1)
        self._data.extend(new_data)
        self.endInsertRows()

    def clear_data(self):
        self.beginRemoveRows(QModelIndex(), 0, self.rowCount() - 1)
        self._data.clear()
        self.endRemoveRows()

    def get_data(self):
        return self._data

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = None):
        if role == Qt.DisplayRole:
            return self._header[section]

    def data(self, index: QModelIndex, role: int = None):
        if role == Qt.DisplayRole:
            return self._data[index.row()][index.column()]

    def rowCount(self, parent: QModelIndex = None) -> int:
        return len(self._data)

    def columnCount(self, parent: QModelIndex = None) -> int:
        return len(self._header)


class UploaderBackend(QObject):
    def __init__(self, provider: YtmusicProvider = None):
        super(UploaderBackend, self).__init__()
        self._provider = provider
        self._upload_thread: Optional[threading.Thread] = None
        self._thread_paused = False
        self._uploading_song_model = UploadingSongListModel(['文件路径', '上传状态'], [])

    @pyqtProperty(UploadingSongListModel, constant=True)
    def uploading_song_model(self) -> UploadingSongListModel:
        return self._uploading_song_model

    @pyqtSlot(name='clearAll')
    def clear_all(self):
        self._uploading_song_model.clear_data()

    def _update_upload_state(self, index: int, state: bool):
        self._uploading_song_model.update_state(index, state)

    def _upload(self):
        for i, song in enumerate(self._uploading_song_model.get_data()):
            file = Path(song[0])
            if not file.exists():
                continue
            state = self._provider.upload_song(file.as_posix())
            self._update_upload_state(i, state)
            while self._thread_paused:
                time.sleep(1)
        self._upload_thread = None

    def __del__(self):
        if self._upload_thread is not None:
            del self._upload_thread
            self._upload_thread = None

    @pyqtProperty(str, constant=True)
    def is_uploading(self) -> str:
        if self._upload_thread is None:
            return '上传'
        elif self._thread_paused:
            return '继续'
        else:
            return '暂停'

    @pyqtSlot()
    def upload(self):
        if self._upload_thread is not None:
            self._thread_paused = False
            return
        self._upload_thread = threading.Thread(target=self._upload)
        self._upload_thread.start()

    @pyqtSlot()
    def pause(self):
        if self._upload_thread is not None:
            return
        self._thread_paused = True

    @pyqtSlot(QVariant, name='filesDropped')
    def files_dropped(self, value: QVariant) -> bool:
        if len(value) == 0:
            return False
        files_data = []
        # noinspection PyTypeChecker
        for url in value:
            url: QUrl
            path = Path(url.toLocalFile())
            if path.is_file():
                files_data.append([path.as_posix(), '待上传'])
            elif path.is_dir():
                # 支持的文件类型 *.mp3 *.m4a, *.wma, *.flac, *.ogg
                extensions = ['*.mp3', '*.m4a', '*.wma', '*.flac', '*.ogg', '*.MP3', '*.M4A', '*.WMA', '*.FLAC', '*.OGG']
                for e_ in extensions:
                    p_ = path.rglob(e_)
                    for p in p_:
                        files_data.append([p.as_posix(), '待上传'])
        if len(files_data) == 0:
            return False
        self._uploading_song_model.insert_data(files_data)
        return True


async def render(req, **kwargs):
    app = req.ctx['app']
    app.ui.right_panel.set_body(app.ui.table_container)

    provider: YtmusicProvider = app.library.get('ytmusic')
    tab_id = Tab(int(req.query.get('tab_id', Tab.songs.value)))
    renderer = FavRenderer(tab_id, provider)
    await app.ui.table_container.set_renderer(renderer)


class FavRenderer(Renderer, LibraryTabRendererMixin):
    def __init__(self, tab_id, provider):
        self.tab_id = tab_id
        self._provider: YtmusicProvider = provider
        self._backend = UploaderBackend(provider)
        self._uploader: Optional[QQuickView] = None
        self._load_uploader()

    def _load_uploader(self):
        self._uploader = QQuickView()
        self._uploader.setTitle('上传音乐')
        self._uploader.rootContext().setContextProperty('backend', self._backend)
        self._uploader.setSource(QUrl.fromLocalFile((Path(__file__).parent / 'qml' / 'uploader.qml').as_posix()))
        self._uploader.setResizeMode(QQuickView.SizeRootObjectToView)

    async def render(self):
        self.render_tabbar()
        self.meta_widget.show()
        self.meta_widget.title = '上传的音乐'

        if self.tab_id == Tab.songs:
            self.songs_table: SongsTableView
            self.songs_table.remove_song_func = self._delete_song
            self.show_songs(await aio.run_fn(lambda: self._provider.library_upload_songs()))
            self.toolbar.show()
            # self.toolbar.manual_mode()
            btn = TextButton('上传音乐', self.toolbar)
            btn.clicked.connect(self._show_uploader)
            self.toolbar.add_tmp_button(btn)
        elif self.tab_id == Tab.artists:
            self.show_artists(await aio.run_fn(lambda: self._provider.library_upload_artists()))
        elif self.tab_id == Tab.albums:
            self.show_albums(await aio.run_fn(lambda: self._provider.library_upload_albums()))

    def _delete_song(self, song):
        entity_id = song.entityId
        ok = self._provider.delete_uploaded_song(entity_id)
        if not ok:
            QMessageBox.warning(self.toolbar, '删除歌曲', '删除失败！')
        else:
            QMessageBox.information(self.toolbar, '删除歌曲', '删除成功！')
        return ok

    def _show_uploader(self):
        self._uploader.show()

    def _upload_song(self):
        path, _ = QFileDialog.getOpenFileName(self.toolbar, '选择文件', Path.home().as_posix(),
                                              'Supported Files (*.mp3 *.m4a, *.wma, *.flac, *.ogg);; All Files (*.*)')
        if path == '':
            return
        ok = self._provider.upload_song(path)
        if not ok:
            QMessageBox.warning(self.toolbar, '上传音乐', '上传失败！')
        else:
            QMessageBox.information(self.toolbar, '上传音乐', '上传成功！')

    def show_by_tab_id(self, tab_id):
        query = {'tab_id': tab_id.value}
        self._app.browser.goto(page='/providers/ytmusic/more', query=query)

    def render_tabbar(self):
        super().render_tabbar()
        try:
            self.tabbar.songs_btn.setText('上传的歌曲')
            self.tabbar.albums_btn.setText('上传的专辑')
            self.tabbar.artists_btn.setText('上传的艺术家')
            self.tabbar.videos_btn.hide()
            self.tabbar.playlists_btn.hide()
        except Exception as e:
            logger.warning(str(e))
